from pyanten.Utility.frequencyToWavelength import frequencyToWavelength

def wavelengthToFrequency(wavelength: float):
  """ Converts frequency to wavelength

  Parameters
  ----------
  wavelength : float
      wavelength in meters [m]

  Returns
  -------
  frequency : float
      frequency in Hertz [Hz]
  """

  return frequencyToWavelength(wavelength)


from math import log10

from pyanten.Utility.swrToGamma import swrToGamma

def swrToGammaDb(swr: float):
  """ Voltage standing wave ratio (vswr) to reflection coefficient (gamma) in decibel conversion

  Parameters
  ----------
  swr : float
      standing wave ratio (swr)

  Returns
  -------
  gamma : float
      reflection coefficient (s11) [dB]
  """
  
  return 20*log10(swrToGamma(swr))


from math import floor, ceil

from pyanten.Measurement.NearField.Planar.samplingLength import samplingLength

def samplingCount(frequency: float, L: float):
  r""" Planar near-field antenna measurement sampling count according to frequency and desired minimum length

  Parameters
  ----------
  frequency : float
              frequency of interest in Hertz [Hz]
  L         : float
              desired minimum scan length [m]

  Returns
  -------
  Lm    : float
          sampling start position [m]
  Lp    : float
          sampling stop position [m] (equals to negative of Lp)
  N     : float
          samplng count
  Delta : float
          spatial samplng length [mm]

  Notes
  -----
  Calculation assumes a scanning region centered on the AUT.  

  Formula
  -------
  .. math::
      theta = \tan^{-1} \left \dfrac{L-a}{2d} \right
  """
  
  samplingLengthInmm = floor(samplingLength(frequency)*1E3)
  scanLengthInmm     = ceil(L*1E3)
  halfScanLengthInmm = scanLengthInmm/2 if scanLengthInmm%2 ==0 else (scanLengthInmm+1)/2
  if halfScanLengthInmm%samplingLengthInmm > 0:
    halfScanLengthInmm = halfScanLengthInmm + (samplingLengthInmm-halfScanLengthInmm%samplingLengthInmm)
  
  N  = 2*(halfScanLengthInmm/samplingLengthInmm)+1
  Lm = -halfScanLengthInmm
  Lp = +halfScanLengthInmm
  
  return ( Lm/1E3, Lp/1E3, int(N), int(samplingLengthInmm) )


from math import log10

def pathLoss(frequency: float, R: float, G1: float, G2:float):
  r""" Free space path loss under far-field condition

  Parameters
  ----------
  frequency : float
              frequency of interest in MegaHertz [MHz]
  R         : float
              range in kilometers [km]
  G1        : float
              gain of transmitter antenna in decibels [dB]
  G2        : float
              gain of receiver antenna in decibels [dB]

  Returns
  -------
  Lp : float
       path loss in decibels [dB]

  Notes
  -----
  This calculation valid if two antennas are seperated enough to be inside far-field (Fraunhofer) region.

  References
  ----------
  .. [1] IEEE 1720-2012 Recommended Practice for Near-Field Antenna Measurements, Section B.2.2, Page 175, Equation B.6.
  .. [2] Warren L. Stutzman, Antenna Theory and Design, 3rd Ed., Page 43.

  Formula
  -------
  .. math::
      L_P = 32.45 + 20\log10(fR) - G_1 - G_2 \quad [dB]
  """

  return 32.45 + 20*log10(frequency*R) - G1 - G2



import math

def gammaDbToReflectedPowerPercentage(gamma: float) -> float:
  """ Reflection coefficient (gamma) in decibels to reflected power percentage

  Parameters
  ----------
  gamma : float
    reflection coefficient (s11) [dB]

  Returns
  -------
  float
    reflected power percentage
  """

  return math.pow(10,gamma/20)*math.pow(10,gamma/20)*100


from pyanten.Utility import frequencyToWavelength

def samplingLength(frequency: float):
  r""" Maximum sampling length for near-field antenna measurements

  Parameters
  ----------
  frequency : float
      frequency of interest in Hertz [Hz]

  References
  -----
  .. [1] IEEE 149-2021 Recommended Practice for Antenna Measurements, Section 12.5, Page 134.
  .. [2] IEEE 1720-2012 Recommended Practice for Near-Field Antenna Measurements, Section 5.2.5, Page 23, Equation 25.

  Formula
  -------
  .. math::
      \Delta = \lambda / 2 = 0.5 \times (c_0 / f)
  """
  return 0.5*frequencyToWavelength(frequency)


from enum import Enum
from scipy.signal.windows import taylor, chebwin


def amplitudeTaper(N: int, sll: float, window: str = "taylor", **kwargs):
  r""" Generate amplitude taper coefficients for antenna elements in sidelobe controlle array 

  Parameters
  ----------
  N      : int 
           number of array elements
  sll    : float
           desired sidelobe level [dB]
  taper  : str 
           window function associated to desired distribution that can be:
           taylor,
           dolhp-chebyshev / chebyshev / tchebyshev 
**kwargs : dict
           window specific properties. Valid keyword arguments are: 

           taylor_nbar: int: Number of nearly constant-level sidelobes adjacent to the mainlobe in taylor window (default=4)
         
  Returns
  -------
  amplitudes : float[]
               amplitude array corresponds to each element

  References
  ----------
  .. [1] [taylor — SciPy v1.14.1 Manual](https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.taylor.html)
  .. [1] [chebwin — SciPy v1.14.1 Manual](https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.chebwin.html)
  .. [2] [Armin Doerry, “Catalog of Window Taper Functions for Sidelobe Control”, 2017](https://www.osti.gov/biblio/1365510)
  .. [3] [Sayedu Khasim, Noorbasha & Krishna Y, Murali & Thati, Jagadeesh & Subbarao, M.. (2013). ANALYSIS OF DIFFERENT TAPERING TECHNIQUES FOR EFFICIENT RADIATION PATTERN. e-Journal of Science & Technology (e-JST). 8. 47-55. ](https://www.dl.begellhouse.com/journals/0632a9d54950b268,4b96ebea37254651,10b28d794a40eb40.html)
  """

  if window == "dolhp-chebyshev" or window == "chebyshev" or window == "tchebyshev":
     return chebwin(N, at=sll)
  else:
     return taylor(N, nbar = 4 if "taylor_nbar" not in kwargs.keys() else kwargs["taylor_nbar"], sll=sll)
  





from math import sqrt

def lineOfSight(height: float):
  r""" Line-of-sight range in kilometers [km] for a given height

  Parameters
  ----------
  height : float
           height of transmitter/recevier above ground in meters [m]

  Returns
  -------
  LOS : float
        line-of-sight distance in kilometers [km]

  References
  ----------
  .. [1] [Line-of-sight propagation - Wikipedia](https://en.wikipedia.org/wiki/Line-of-sight_propagation)

  Formula
  -------
  .. math::
      R_{los} = \sqrt{2\cdotR\cdoth} \approx 3.57\times\sqrt{h} [km]
  """

  return 3.57*sqrt(height)


from math import tan, radians

def scanLength(a: float, d: float, theta:float):
  r""" Required length of the scan for desired angle-of-view in planar near-field antenna measurements

  Parameters
  ----------
  a     : float
          antenna cross-section length
  d     : float
          seperation distance between antenna and probe
  theta : float
          desired pattern view angle along one side [deg]

  Notes
  -----
  Calculation assumes a scanning region centered on the AUT.
  Input distances are unitless so output is the same quantity of inputs.

  References
  ----------
  .. [1] IEEE 149-2021 Recommended Practice for Antenna Measurements, Section 12.5, Page 135, Equation 99.
  .. [2] IEEE 1720-2012 Recommended Practice for Near-Field Antenna Measurements, Section 5.3.1.6, Page 28, Equation 27.

  Formula
  -------
  .. math::
      L = 2d \cdot \tan\theta + a
  """
  
  return 2*d * tan(radians(theta)) + a



def gammaToSwr(gamma: float):
  """ Reflection coefficient (gamma) to voltage standing wave ratio (vswr) conversion

  Parameters
  ----------
  gamma : float
      reflection coefficient (s11)

  Returns
  -------
  swr : float
      standing wave ratio (swr)
  """
  return (1+gamma)/(1-gamma)


from math import sqrt

def radioHorizon(height: float):
  r""" Radio horizon in kilometers [km] for a given height

  Parameters
  ----------
  height : float
           height of transmitter/recevier above ground in meters [m]

  Returns
  -------
  Rh : float
       radio horizon in kilometers [km]

  References
  ----------
  .. [1] [Line-of-sight propagation - Wikipedia](https://en.wikipedia.org/wiki/Line-of-sight_propagation)

  Formula
  -------
  .. math::
      R_{rh} \approx 4.12\times\sqrt{h} [km]
  """
  
  return 4.12*sqrt(height)


from .frequencyToWavelength import frequencyToWavelength
from .wavelengthToFrequency import wavelengthToFrequency
from .swrToGamma import swrToGamma
from .swrToGammaDb import swrToGammaDb
from .gammaToSwr import gammaToSwr
from .gammaDbToSwr import gammaDbToSwr
from .gammaDbToReflectedPowerPercentage import gammaDbToReflectedPowerPercentage


from pyanten.Utility import frequencyToWavelength

def farFieldDistance(frequency: float, D: float):
  r""" Far-field (Fraunhofer) distance of an antenna or aperture

  Parameters
  ----------
  frequency : float
              frequency of interest in Hertz [Hz]
  D         : float
              maximum cross-sectional size [m]

  Returns
  -------
  Rff : float
        minimum far-field range in meters [m]

  References
  ----------
  .. [1] IEEE 1720-2012 Recommended Practice for Near-Field Antenna Measurements, Section B.2.2, Page 175, Equation B.6.
  .. [2] Warren L. Stutzman, Antenna Theory and Design, 3rd Ed., Page 43.

  Formula
  -------
  .. math::
      R_{ff} = Max\left(\dfrac{2D^2}{\lambda},5D,1.6\lambda\right)
  """

  wavelength = frequencyToWavelength(frequency)
  return max([2*D*D/wavelength, 5*D, 1.6*wavelength])






from math import pow

from pyanten.Utility.gammaToSwr import gammaToSwr

def gammaDbToSwr(gamma: float):
  """ Reflection coefficient (gamma) in decibels to voltage standing wave ratio (vswr) conversion

  Parameters
  ----------
  gamma : float
      reflection coefficient (s11) [dB]

  Returns
  -------
  swr : float
      standing wave ratio (swr)
  """

  return gammaToSwr(pow(10,gamma/20))


from math import atan2, degrees

def angleOfView(a: float, d: float, L: float):
  r""" Reliable far-field angle-of-view in planar near-field antenna measurements

  Parameters
  ----------
  a : float
      antenna cross-section length
  d : float
      seperation distance between antenna and probe
  L : float
      scan length of region

  Notes
  -----
  Calculation assumes a scanning region centered on the AUT.  
  Input distances are unitless but should in same quantity

  References
  ----------
  .. [1] IEEE 149-2021 Recommended Practice for Antenna Measurements, Section 12.5, Page 135, Equation 99.
  .. [2] IEEE 1720-2012 Recommended Practice for Near-Field Antenna Measurements, Section 5.3.1.6, Page 28, Equation 27.

  Formula
  -------
  .. math::
      theta = \tan^{-1} \left \dfrac{L-a}{2d} \right
  """

  return degrees(atan2(L-a, 2*d))


from pyanten.Utility import frequencyToWavelength

def minimumRange(frequency: float, closer: bool = False) -> float:
  r""" Minimum recommended distance for near-field antenna measurements in meters

  Parameters
  ----------
  frequency : float
    frequency of interest in Hertz [Hz]

  closer : bool
    closer approximation using 3 wavelength calculation

  Returns
  -------
  float
    minimum distance

  Notes
  -----
  Standard suggests to choose between 3 or 5 wavelength.
  In order to ensure copuling effect, 5 wavelength distance is choosen and implemented. 3 wavelength is optional.

  .. [1] IEEE 1720-2012 Recommended Practice for Near-Field Antenna Measurements, Section 5.3.1.4, Page 27.

  Formula
  -------
  .. math::
      R_{nf} = 5\times\lambda = 5\times c_0/f
  """
  return 3*frequencyToWavelength(frequency) if closer else 5*frequencyToWavelength(frequency)


from math import sqrt

def maximumRadiusExtent(dx: float, dy:float, dz:float):
  r""" Calculates maximum radial extend (MRE) of the antenna for spherical near-field (SNF) measurements

  Parameters
  ----------
  dx : float
       max displacement from rotation center along x-axis
  dy : float
       max displacement from rotation center along y-axis
  dz : float
       max displacement from rotation center along z-axis

  Returns
  -------
  mre : float
        maximum radius extent

  Notes
  -----
  Calculation is unitless so output is the same quantity of inputs.

  Formula
  -------
  .. math::
      mre = \sqrt{dx^2+dy^2+dz^2}
  """
  
  return sqrt(dx*dx + dy*dy + dz*dz)




def swrToGamma(swr: float):
  """ Voltage standing wave ratio (vswr) to reflection coefficient (gamma) conversion

  Parameters
  ----------
  swr : float
      standing wave ratio (swr)

  Returns
  -------
  gamma : float
      reflection coefficient (s11)
  """
  return (swr-1)/(swr+1)


from scipy.constants import speed_of_light

def frequencyToWavelength(frequency: float):
  """ Converts frequency to wavelength

  Parameters
  ----------
  frequency : float
      frequency in Hertz [Hz]

  Returns
  -------
  wavelength : float
      wavelength in meters [m]
  """
  return speed_of_light / frequency